import * as THREE from '../build/three.module.js';
import Stats from './jsm/libs/stats.module.js';
import { GUI } from './jsm/libs/dat.gui.module.js';
import { OrbitControls } from './jsm/controls/OrbitControls.js';
import { BufferGeometryUtils } from './jsm/utils/BufferGeometryUtils.js';

const Method = {
    INSTANCED: "INSTANCED",
    MERGED: "MERGED",
    NAIVE: "NAIVE"
};

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        const domWebGL = document.createElement('div');
        document.body.appendChild(domWebGL);
        this._domWebGL = domWebGL;

        const renderer = new THREE.WebGLRenderer({ antialias: true });
        domWebGL.appendChild(renderer.domElement);
        renderer.outputEncoding = THREE.sRGBEncoding; // 유무차이 
        this._renderer = renderer;

        const scene = new THREE.Scene();
        scene.background = new THREE.Color(0xffffff);
        this._scene = scene;

        this._setupCamera();
        this._setupControls();
        this._setupGUI();

        //this._setupLight();
        this._setupModel();
        

        //this._setupHelper();
        
        window.onresize = this.resize.bind(this);

        this.resize();
    }

    _setupModel() {
        this._initMesh();
    }

    _setupHelper() {
        //.
    }

    _clean() {
        let cntMeshs = 0;
        const meshes = [];
        this._scene.traverse((object) => {
            if(object.isMesh) {
                meshes.push(object);
                cntMeshs++;
            }
        })

        console.log("Mesh Count: " + cntMeshs);

        for(let i=0; i<meshes.length; i++) {
            const mesh = meshes[i];
            mesh.material.dispose();
            mesh.geometry.dispose();
            this._scene.remove(mesh);
        }
    }

    _initMesh() {
        this._clean();

        new THREE.BufferGeometryLoader()
            .setPath("models/json/")
            .load("suzanne_buffergeometry.json", geometry => {
                const material = new THREE.MeshNormalMaterial();
                geometry.computeVertexNormals();
                console.time(this._api.method + " (build)");

                switch(this._api.method) {
                    case Method.INSTANCED:
                        this._makeInstanced(geometry, material);
                        break;
                    case Method.MERGED:
                        this._makeMerged(geometry, material);
                        break;
                    case Method.NAIVE:
                        this._makeNaive(geometry, material);
                        break;
                }

                console.timeEnd(this._api.method + " (build)");
            });
    }

    _randomizeMatrix(matrix) {
        const position = new THREE.Vector3();
        const rotation = new THREE.Euler();
        const quaternion = new THREE.Quaternion();
        const scale = new THREE.Vector3();

        position.x = Math.random() * 40 - 20;
        position.y = Math.random() * 40 - 20;
        position.z = Math.random() * 40 - 20;

        rotation.x = Math.random() * 2 * Math.PI;
        rotation.y = Math.random() * 2 * Math.PI;
        rotation.z = Math.random() * 2 * Math.PI;

        quaternion.setFromEuler(rotation);

        scale.x = scale.y = scale.z = Math.random() * 1;

        matrix.compose(position, quaternion, scale);
    }

    _getGeometryByteLength(geometry) {
        let total = 0;
        if(geometry.index) total += geometry.index.array.byteLength;
        for(const name in geometry.attributes) {
            total += geometry.attributes[name].array.byteLength;
        }
        return total;
    }

    // Source: https://stackoverflow.com/a/18650828/1314762
    _formatBytes(bytes, decimals) {
        if (bytes === 0) return '0 bytes';

        const k = 1024;
        const dm = decimals < 0 ? 0 : decimals;
        const sizes = ['bytes', 'KB', 'MB'];

        const i = Math.floor(Math.log(bytes) / Math.log(k));

        return parseFloat((bytes / Math.pow(k, i)).toFixed(dm)) + ' ' + sizes[i];
    }

    _makeNaive(geometry, material) {
        const matrix = new THREE.Matrix4();

        for(let i=0; i<this._api.count; i++) {
            this._randomizeMatrix(matrix);

            const mesh = new THREE.Mesh(geometry, material);
            mesh.applyMatrix4(matrix);

            this._scene.add(mesh);
        }

        this._guiStatsEl.innerHTML = [
            '<i>GPU draw calls</i>: ' + this._api.count,
            '<i>GPU memory</i>: ' + this._formatBytes(this._api.count*16 + this._getGeometryByteLength(geometry), 2)
        ].join('<br/>');
    }

    _makeMerged(geometry, material) {
        const geomtries = [];
        const matrix = new THREE.Matrix4();

        for(let i=0; i<this._api.count; i++) {
            this._randomizeMatrix(matrix);
            const instanceGeometry = geometry.clone();
            instanceGeometry.applyMatrix4(matrix);
            geomtries.push(instanceGeometry);
        }

        const mergedGeometry = BufferGeometryUtils.mergeBufferGeometries(geomtries);
        this._scene.add(new THREE.Mesh(mergedGeometry, material));

        this._guiStatsEl.innerHTML = [
            '<i>GPU draw calls</i>: 1',
            '<i>GPU memory</i>: ' + this._formatBytes(this._getGeometryByteLength(mergedGeometry), 2)
        ].join('<br/>');
    }

    _makeInstanced(geometry, material) {
        const matrix = new THREE.Matrix4();
        const mesh = new THREE.InstancedMesh(geometry, material, this._api.count);

        for(let i=0; i<this._api.count; i++) {
            this._randomizeMatrix(matrix);
            mesh.setMatrixAt(i, matrix);
        }

        this._scene.add(mesh);

        this._guiStatsEl.innerHTML = [
            '<i>GPU draw calls</i>: 1',
            '<i>GPU memory</i>: ' + this._formatBytes(this._api.count*16 + this._getGeometryByteLength(geometry), 2)
        ].join('<br/>');
    }

    _setupGUI() {
        const stats = new Stats();
        this._domWebGL.appendChild(stats.dom);
        this._stats = stats;

        const api = {
            method: Method.INSTANCED,
            count: 1000
        };

        const gui = new GUI();
        gui.add(api, "method", Method).onChange(this._initMesh.bind(this));
        gui.add(api, "count", 1, 100000).step(1).onChange(this._initMesh.bind(this));

        const prefFolder = gui.addFolder("Performance");
        const guiStatsEl = document.createElement("li");
        guiStatsEl.style.height = "auto";

        prefFolder.__ul.appendChild(guiStatsEl);
        prefFolder.open();

        this._api = api;
        this._guiStatsEl = guiStatsEl;
    }

    _setupCamera() {
        const camera = new THREE.PerspectiveCamera(70, window.innerWidth / window.innerHeight, 1, 100);
        camera.position.z = 30;

        this._camera = camera;
    }

    _setupControls() {
        const controls = new OrbitControls(this._camera, this._renderer.domElement);
        controls.autoRotate = true;
    }

    _setupLight() {
        //.
    }

    update() {
        this._stats.update();
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        
        this.update();
        this._renderer.render(this._scene, this._camera);
    }

    resize() {
        const aspect = window.innerWidth / window.innerHeight;

        this._camera.aspect = aspect;
        this._camera.updateProjectionMatrix();

        this._renderer.setPixelRatio( window.devicePixelRatio );
        this._renderer.setSize(window.innerWidth, window.innerHeight);
    }
}

window.onload = function () {
    (new App()).render();
}